from typing import Dict, Any, List, Optional

from spotlight.api.alert.__util import (
    _get_alert_request_info,
    _get_alerts_request_info,
    _create_alert_request_info,
    _update_alert_request_info,
    _delete_alert_request_info,
    _get_alert_signal_request_info,
)
from spotlight.api.alert.model import AlertRequest
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import (
    __async_get_request,
    __async_post_request,
    __async_put_request,
    __async_delete_request,
)


@async_data_request()
async def async_get_alert(id: str) -> Dict[str, Any]:
    """
    Asynchronously get alert by ID.

    Args:
        id (str): Alert ID

    Returns:
        Dict[str, Any]: Alert response
    """
    request_info = _get_alert_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request()
async def async_get_alerts(
    override_find_all: Optional[bool] = False,
) -> List[Dict[str, Any]]:
    """
    Asynchronously get all alerts.

    Args:
        override_find_all (Optional[bool]): If True, overrides the default limitation to fetch all results. This parameter is restricted to admin use only. Non-admin users will receive a 'Forbidden' exception if they attempt to use it. Defaults to False.

    Returns:
        List[Dict[str, Any]]: List of alert responses
    """
    request_info = _get_alerts_request_info(override_find_all)
    return await __async_get_request(**request_info)


@async_data_request()
async def async_get_alert_signal(start: int, end: int) -> List[Dict[str, Any]]:
    """
    Asynchronously get signals for each alert in time window.

    Args:
        start (int): Unix timestamp for the start of the time window
        end (int): Unix timestamp for the end of the time window

    Returns:
        Dict[str, Any]: AlertSignal response
    """
    request_info = _get_alert_signal_request_info(start, end)
    return __async_get_request(**request_info)


@async_data_request()
async def async_create_alert(request: AlertRequest) -> Dict[str, Any]:
    """
    Asynchronously create alert.

    Args:
        request (AlertRequest): Alert request

    Returns:
        Dict[str, Any]: Alert response
    """
    request_info = _create_alert_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request()
async def async_update_alert(id: str, request: AlertRequest) -> Dict[str, Any]:
    """
    Asynchronously update alert.

    Args:
        id (str): Alert ID
        request (AlertRequest): Alert request

    Returns:
        Dict[str, Any]: Alert response
    """
    request_info = _update_alert_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(processor=lambda response: None)
async def async_delete_alert(id: str) -> None:
    """
    Asynchronously delete alert by ID.

    Args:
        id (str): Alert ID

    Returns:
        None
    """
    request_info = _delete_alert_request_info(id)
    return await __async_delete_request(**request_info)



import logging
import time
from dataclasses import dataclass

logger = logging.getLogger(__name__)


@dataclass
class TimeProfile:
    start_time: float
    end_time: float
    duration: float


def timeit(fxn, *, return_time_profile: bool = False):
    def timed(*args, **kwargs):
        start = time.time()
        result = fxn(*args, **kwargs)
        end = time.time()

        time_profile = TimeProfile(
            start_time=start, end_time=end, duration=(end - start)
        )
        logger.debug(f"func:{fxn.__name__} args:[{args}, {kwargs}]: {time_profile}")

        if return_time_profile:
            return result, time_profile
        return result

    return timed



import asyncio
import random


def async_exponential_backoff_retry(func, max_retries=5, delay=1):
    async def wrapper(*args, **kwargs):
        for attempt in range(max_retries):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                if attempt < max_retries - 1:
                    sleep_time = delay * (2**attempt) + random.uniform(0, 1)
                    await asyncio.sleep(sleep_time)
                else:
                    raise e

    return wrapper



"""
Permission API.
"""

from spotlight.api.permission.asynchronous import (
    async_get_permission,
    async_get_permission_by_resource_id,
    async_get_permissions_by_resource_id,
    async_create_permission,
    async_update_permission,
    async_delete_permission,
)
from spotlight.api.permission.synchronous import (
    get_permission,
    get_permission_by_resource_id,
    get_permissions_by_resource_id,
    create_permission,
    update_permission,
    delete_permission,
)



from typing import Optional

from spotlight.core.common.config import EnvironmentConfig

config = EnvironmentConfig()


def _login_request_info(
    username: str, password: str, scope: Optional[str] = None
) -> dict:
    data = {
        "client_id": "spotlight-api",
        "grant_type": "password",
        "username": username,
        "password": password,
    }
    if scope is not None:
        data.update({"scope": scope})

    return {
        "url": f"{config.get_url('keycloak')}/protocol/openid-connect/token",
        "data": data,
    }


def _refresh_token_request_info(refresh_token: Optional[str] = None) -> dict:
    return {
        "url": f"{config.get_url('keycloak')}/protocol/openid-connect/token",
        "data": {
            "client_id": "spotlight-api",
            "grant_type": "refresh_token",
            "refresh_token": refresh_token,
        },
    }


def _token_exchange_request_info(requested_subject: str, subject_token: str) -> dict:
    data = {
        "client_id": "spotlight-api",
        "grant_type": "urn:ietf:params:oauth:grant-type:token-exchange",
        "requested_subject": requested_subject,
        "subject_token": subject_token,
        "scope": "offline_access",
    }

    return {
        "url": f"{config.get_url('keycloak')}/protocol/openid-connect/token",
        "data": data,
    }



"""
Helper functions.
"""

from functools import reduce
from typing import Dict, Any, Union, List, Optional, Callable

import pandas as pd
from aiohttp import ClientResponse
from pandas import DataFrame
from requests import Response

from spotlight.core.common.errors import ApiError


def coalesce(*args):
    """
    Coalesce operator returns the first arg that isn't None.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    return reduce(lambda x, y: x if x is not None else y, args)


def coalesce_callables(*args):
    """
    Coalesce operator returns the first arg that isn't None. If the arg is callable it will check that the value
    returned from calling the arg is not none and then return the value from calling it.

    WARNING: If an argument implements __call__ this method will evaluate the return of the __call__ method and return
    that instead of the argument itself. This is important when using python classes.

    Args:
        *args: Tuple of args passed to the function

    Returns:
        The first non-None value in *args
    """
    for arg in args:
        value = arg() if callable(arg) else arg
        if value is not None:
            return value
    return None


def flat_map(fxn: Callable[[Any], Any], lst: List) -> List:
    """
    The function maps a function onto a nested iterable and returns a flattened iterable

    Parameters:
        fxn (Callable[[Any], Any])
        lst (List): List for the function to be called on

    Returns:
        (List) A flattened list with the original lists elements transformed via the fxn
    """
    if len(lst) <= 0:
        return []
    return reduce(lambda a, b: a + b, map(fxn, lst))


def transform_one_or_many(
    data: Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]],
    key: Optional[str] = None,
) -> Union[DataFrame, Dict[str, DataFrame]]:
    """
    Converts data to a dataframe or multiple dataframes.

    Args:
        data (Union[List[Dict[str, Any]], Dict[str, List[Dict[str, Any]]]]): Dict being transformed into Dataframes
        key (Optional[str]): Key for indexing the values of the data dict

    Returns:
        Union[DataFrame, Dict[str, DataFrame]]: A single dataframe or a map of names to dataframes
    """
    if isinstance(data, dict):
        for k in data.keys():
            data[k] = DataFrame(data[k][key])
        return data

    return DataFrame(data)


def to_pandas_with_index(data: Dict[str, Any], index: str = "date") -> DataFrame:
    """
    Transform input data into pandas dataframe and assign index.

    Args:
        data (Dict[str, Any]): Input data as dict
        index (str): Index column name

    Returns:
        DataFrame: DataFrame
    """
    # necessary to drop column in order to avoid duplicates when converting to json
    return pd.DataFrame(data).set_index(index, drop=True)


def no_transform(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Passthrough data without any transformation.

    Args:
        data (Dict[str, Any]): Input data

    Returns:
        Dict[str, Any]: Output data
    """
    return data


def data_required_transform(data: Dict[str, Any]) -> Dict[str, Any]:
    """
    Passthrough data without any transformation.

    Args:
        data (Dict[str, Any]): Input data

    Returns:
        Dict[str, Any]: Output data
    """
    if len(data) <= 0:
        raise ApiError(
            f"There are no results for the data you are querying at this time"
        )

    return data


async def build_response_obj(aio_response: ClientResponse) -> Response:
    """
    Construct a requests Response object from an aiohttp Client Response object

    Args:
        aio_response (ClientResponse): Response object from aiohttp

    Returns:
        Response: A requests Response object
    """
    content = await aio_response.content.read()

    response = Response()
    response._content = content
    response.url = str(aio_response.url)
    response.status_code = aio_response.status
    headers = {row[0]: row[1] for row in aio_response.headers.items()}
    response.headers = headers
    return response


def to_nested_dict(data, separator="."):
    """
    This method takes a dict and splits keys by the seperator to create a nested dict

    Parameters:
        data: dict to unwind into a nested dict
        separator: seperator used to split keys to build the nested dict

    Returns:
        A nested dict
    """
    nested_dict = {}
    for key, value in data.items():
        keys = key.split(separator)
        d = nested_dict
        for subkey in keys[:-1]:
            if subkey not in d:
                d[subkey] = {}
            d = d[subkey]
        d[keys[-1]] = value
    return nested_dict



from typing import List, Dict, Optional


def _get_client_id_request_info() -> dict:
    return {
        "endpoint": "clients?clientId=realm-management",
        "url_key": "keycloak_admin",
    }


def _get_users_request_info(limit: int, page_offset: int) -> dict:
    return {
        "endpoint": "users",
        "params": {"max": limit, "first": page_offset},
        "url_key": "keycloak_admin",
    }


def _get_user_count_request_info() -> dict:
    return {"endpoint": f"users/count", "url_key": "keycloak_admin"}


def _get_user_request_info(user_id: str) -> dict:
    return {"endpoint": f"users/{user_id}", "url_key": "keycloak_admin"}


def _execute_actions_email_request_info(
    user_id: str, actions: List[str], client_id: str, redirect_uri: str
) -> dict:
    return {
        "endpoint": f"users/{user_id}/execute-actions-email",
        "params": {
            "client_id": client_id,
            "redirect_uri": redirect_uri,
            "lifespan": 604800,
        },
        "url_key": "keycloak_admin",
        "json": actions,
    }


def _update_user_request_info(user_id: str, data: dict) -> dict:
    return {"endpoint": f"users/{user_id}", "url_key": "keycloak_admin", "json": data}


def _get_groups_request_info() -> dict:
    return {"endpoint": "groups", "url_key": "keycloak_admin"}


def _get_group_members_request_info(id: str) -> dict:
    return {"endpoint": f"groups/{id}/members", "url_key": "keycloak_admin"}


def _add_group_to_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin",
    }


def _delete_group_from_user_request_info(user_id: str, group_id: str) -> dict:
    return {
        "endpoint": f"users/{user_id}/groups/{group_id}",
        "url_key": "keycloak_admin",
    }


def _get_roles_request_info() -> dict:
    return {"endpoint": "roles", "url_key": "keycloak_admin"}


def _get_roles_for_user_request_info(user_id: str) -> dict:
    return {"endpoint": f"users/{user_id}/role-mappings", "url_key": "keycloak_admin"}


def _add_role_mapping_to_user_request_info(
    user_id: str, role_ids: List[Dict[str, str]]
) -> dict:
    return {
        "endpoint": f"users/{user_id}/role-mappings/realm",
        "url_key": "keycloak_admin",
        "json": role_ids,
    }


def _get_offline_token_request_info(user_id: str) -> dict:
    return {"endpoint": f"config/offline_token/{user_id}"}


def _create_offline_token_request_info(
    user_id: str, refresh_token: str, exp: Optional[int] = None
) -> dict:
    return {
        "endpoint": f"config/offline_token",
        "json": {"user_id": user_id, "refresh_token": refresh_token, "exp": exp},
    }


def _update_offline_token_request_info(
    user_id: str, refresh_token: str, exp: Optional[int] = None
) -> dict:
    return {
        "endpoint": f"config/offline_token/{user_id}",
        "json": {"user_id": user_id, "refresh_token": refresh_token, "exp": exp},
    }


def _delete_offline_token_request_info(user_id: str) -> dict:
    return {"endpoint": f"config/offline_token/{user_id}"}



"""
Admin API.
"""

from spotlight.admin.asynchronous import (
    async_get_client_id,
    async_get_users,
    async_get_user,
    async_update_user,
    async_execute_actions_email,
    async_get_groups,
    async_get_group_members,
    async_add_group_to_user,
    async_delete_group_from_user,
    async_get_roles,
    async_get_roles_for_user,
    async_add_role_mapping_to_user,
    async_get_offline_token,
    async_create_offline_token,
    async_update_offline_token,
    async_delete_offline_token,
    async_get_user_count,
)
from spotlight.admin.synchronous import (
    get_client_id,
    get_users,
    get_user,
    update_user,
    execute_actions_email,
    get_groups,
    get_group_members,
    add_group_to_user,
    delete_group_from_user,
    get_roles,
    get_roles_for_user,
    add_role_mapping_to_user,
    get_offline_token,
    create_offline_token,
    update_offline_token,
    delete_offline_token,
    get_user_count,
)



"""
Authenticated request decorators.
"""

from spotlight.core.common.decorators.authorization.asynchronous import (
    async_authenticated_request,
)
from spotlight.core.common.decorators.authorization.synchronous import (
    authenticated_request,
)



"""
Base asynchronous and synchronous request functions inherited by all API request functions.
"""

from spotlight.core.common.requests.asynchronous import (
    __async_get_request,
    __async_put_request,
    __async_post_request,
    __async_delete_request,
)
from spotlight.core.common.requests.synchronous import (
    __get_request,
    __put_request,
    __post_request,
    __delete_request,
)



from spotlight.api.permission.enum import PermissionType
from spotlight.core.common.base import Base


class PermissionRequest(Base):
    resource_id: str
    permission_type: PermissionType
    permission_id: str
    view: bool
    edit: bool
    delete: bool
    edit_permission: bool



"""
Common types.
"""

from datetime import date, datetime
from typing import NewType, Union

from pandas import DataFrame

Datetime = NewType("Datetime", Union[date, datetime])
DataResponse = NewType("DataResponse", Union[DataFrame, float])



"""
Admin API requests.
"""

from typing import List, Dict, Optional

from spotlight.admin.__util import (
    _get_client_id_request_info,
    _get_users_request_info,
    _get_user_request_info,
    _update_user_request_info,
    _get_groups_request_info,
    _get_group_members_request_info,
    _add_group_to_user_request_info,
    _delete_group_from_user_request_info,
    _get_roles_request_info,
    _get_roles_for_user_request_info,
    _add_role_mapping_to_user_request_info,
    _execute_actions_email_request_info,
    _create_offline_token_request_info,
    _update_offline_token_request_info,
    _get_offline_token_request_info,
    _delete_offline_token_request_info,
    _get_user_count_request_info,
)
from spotlight.core.common.decorators import data_request
from spotlight.core.common.function import no_transform
from spotlight.core.common.requests import (
    __get_request,
    __put_request,
    __post_request,
    __delete_request,
)


@data_request(transformers=[no_transform])
def get_client_id():
    """
    Get keycloak client ID.

    Returns:
        Keycloak client ID.
    """
    request_info = _get_client_id_request_info()
    return __get_request(**request_info)


@data_request
def get_users(limit: int = 500, page_offset: int = 0):
    """
    Get all users.

    Returns:
        List of user response.
    """
    request_info = _get_users_request_info(limit, page_offset)
    return __get_request(**request_info)


@data_request(transformers=[no_transform])
def get_user_count():
    """
    Get the total number of users.

    Returns:
        The total number of users.
    """
    request_info = _get_user_count_request_info()
    return __get_request(**request_info)


@data_request(transformers=[no_transform])
def get_user(user_id: str):
    """
    Get user by ID.

    Args:
        user_id (str): User ID

    Returns:
        User response
    """
    request_info = _get_user_request_info(user_id)
    return __get_request(**request_info)


@data_request(transformers=[no_transform], processor=lambda r: r)
def execute_actions_email(
    user_id: str, actions: List[str], client_id: str, redirect_uri: str
):
    """
    Update user.

    Args:
        user_id (str): User ID
        actions: (List[str]): Required user actions
        client_id: (str): Client ID
        redirect_uri: (str): Redirect URI after executing actions

    Returns:
        User response
    """
    request_info = _execute_actions_email_request_info(
        user_id, actions, client_id, redirect_uri
    )
    return __put_request(**request_info)


@data_request(transformers=[no_transform], processor=lambda r: r)
def update_user(user_id: str, data: dict):
    """
    Update user.

    Args:
        user_id (str): User ID
        data (dict): User request

    Returns:
        User response
    """
    request_info = _update_user_request_info(user_id, data)
    return __put_request(**request_info)


@data_request(transformers=[no_transform])
def get_groups():
    """
    Get groups.

    Returns:
        List of groups.
    """
    request_info = _get_groups_request_info()
    return __get_request(**request_info)


@data_request(transformers=[no_transform])
def get_group_members(id: str):
    """
    Get group members given group ID.

    Args:
        id (str): Group ID

    Returns:
        List of user response
    """
    request_info = _get_group_members_request_info(id)
    return __get_request(**request_info)


@data_request(transformers=[no_transform], processor=lambda r: r)
def add_group_to_user(user_id: str, group_id: str):
    """
    Add user to group.

    Args:
        user_id (str): User ID
        group_id (str): Group ID

    Returns:
        None
    """
    request_info = _add_group_to_user_request_info(user_id, group_id)
    return __put_request(**request_info)


@data_request(transformers=[no_transform], processor=lambda r: r)
def delete_group_from_user(user_id: str, group_id: str):
    """
    Delete user from group.

    Args:
        user_id (str): User ID
        group_id (str): Group ID

    Returns:
        None
    """
    request_info = _delete_group_from_user_request_info(user_id, group_id)
    return __delete_request(**request_info)


@data_request(transformers=[no_transform])
def get_roles():
    """
    Get roles.

    Returns:
        List of role response.
    """
    request_info = _get_roles_request_info()
    return __get_request(**request_info)


@data_request(transformers=[no_transform])
def get_roles_for_user(user_id: str):
    """
    Get roles for user.

    Args:
        user_id (str): User ID

    Returns:
        List of role response
    """
    request_info = _get_roles_for_user_request_info(user_id)
    return __get_request(**request_info)


@data_request(transformers=[no_transform], processor=lambda r: r)
def add_role_mapping_to_user(user_id: str, role_ids: List[Dict[str, str]]):
    """
    Add role mapping to user.

    Args:
        user_id (str): User ID
        role_ids (List[Dict[str, str]]): List of role ID mapping

    Returns:
        None
    """
    request_info = _add_role_mapping_to_user_request_info(user_id, role_ids)
    return __post_request(**request_info)


@data_request(transformers=[no_transform])
def get_offline_token(user_id: str) -> dict:
    """
    Get offline token for user.

    Args:
        user_id (str): User ID

    Returns:
        dict: Offline token response
    """
    request_info = _get_offline_token_request_info(user_id)
    return __get_request(**request_info)


@data_request(transformers=[no_transform])
def create_offline_token(
    user_id: str, refresh_token: str, exp: Optional[int] = None
) -> dict:
    """
    Create offline token for user.

    Args:
        user_id (str): User ID
        refresh_token (str): Refresh token
        exp: (Optional[Int]): Expiration timestamp

    Returns:
        dict: Offline token response
    """
    request_info = _create_offline_token_request_info(user_id, refresh_token, exp)
    return __put_request(**request_info)


@data_request(transformers=[no_transform])
def update_offline_token(
    user_id: str, refresh_token: str, exp: Optional[int] = None
) -> dict:
    """
    Update offline token for user.

    Args:
        user_id (str): User ID
        refresh_token (str): Refresh token
        exp: (Optional[Int]): Expiration timestamp

    Returns:
        dict: Offline token response
    """
    request_info = _update_offline_token_request_info(user_id, refresh_token, exp)
    return __post_request(**request_info)


@data_request(transformers=[no_transform])
def delete_offline_token(user_id: str) -> None:
    """
    Delete offline token for user.

    Args:
        user_id (str): User ID

    Returns:
        None
    """
    request_info = _delete_offline_token_request_info(user_id)
    return __delete_request(**request_info)



"""
Retry decorators.
"""

from spotlight.core.common.decorators.retry.asynchronous import (
    async_exponential_backoff_retry,
)
from spotlight.core.common.decorators.retry.synchronous import exponential_backoff_retry



"""
Base enum class.
"""

from enum import Enum


class BaseEnum(str, Enum):
    """
    Base enum class used by all enum classes.

    Note: Inheriting from str is necessary to correctly serialize output of enum
    """

    pass



def init_subclass(cls):
    cls._subtypes_[cls.__name__] = cls


def parse_obj(cls, obj):
    return cls._convert_to_real_type_(obj)


def get_validators(cls):
    yield cls._convert_to_real_type_


def convert_to_real_type(cls, data):
    if not isinstance(data, dict):
        return data

    data_type = data.get("descriptor")

    if data_type is None:
        raise ValueError("Missing 'descriptor'")

    searched_types = set()

    def search_for_subtype(cls):
        nonlocal searched_types, data_type

        sub = cls._subtypes_.get(data_type)
        if sub:
            return sub

        searched_types.add(cls)
        for _, sub_type in cls._subtypes_.items():
            if sub_type not in searched_types:
                result = search_for_subtype(sub_type)
                if result:
                    return result
        return None

    sub = search_for_subtype(cls)
    if sub is None:
        raise TypeError(f"Unsupport sub-type: {data_type}")

    return sub(**data)



from typing import Callable, Union

from pydantic import Field

from spotlight.core.common.base import Base
from spotlight.core.common.decorators.serializable.__util import (
    parse_obj,
    init_subclass,
    get_validators,
    convert_to_real_type,
)


def serializable(_cls: Base = None) -> Union[Callable, Base]:
    """
    This decorator adds an identifier to the decorated class for identification purposes when serializing and
    deserializing.

    NOTE: This decorator should be used on all children of base abstract classes that use the @serializable_base_class
    decorator.

    Args:
        _cls (object): The class being decorated

    Returns:
        cls: The updated class with the descriptor field
    """

    def wrap(cls):
        return type(cls.cls_name(), (cls,), {"descriptor": Field(cls.cls_name())})

    if _cls is None:
        return wrap

    return wrap(_cls)


def serializable_base_class(_cls: Base = None) -> Union[Callable, Base]:
    """
    This is a decorator sets up functionality to track all subtypes of a parent class. If you want to use a base
    class as a type signature and want the subtypes to be properly deserialized you need to add this decorator to
    the base class and @serializable to all of its children.

    NOTE: You should only use this decorator for base abstract classes

    Args:
        _cls (object): The class being decorated

    Returns:
        cls: The updated class with the descriptor field
    """

    def wrap(cls):
        setattr(cls, "_subtypes_", dict())
        setattr(cls, "parse_obj", classmethod(parse_obj))
        setattr(cls, "__init_subclass__", classmethod(init_subclass))
        setattr(cls, "__get_validators__", classmethod(get_validators))
        setattr(cls, "_convert_to_real_type_", classmethod(convert_to_real_type))
        return cls

    if _cls is None:
        return wrap

    return wrap(_cls)



from typing import Optional

from spotlight.api.data.model import (
    TimeseriesQueryRequest,
    DistinctQueryRequest,
    QueryRequest,
)


def _query_timeseries_request_info(
    request: TimeseriesQueryRequest, one: Optional[bool] = False
) -> dict:
    return {
        "endpoint": f"data/v1.1/timeseries",
        "json": request.request_dict(),
        "params": {"one": True} if one else {},
    }


def _query_timeseries_csv_request_info(
    id: str, request: TimeseriesQueryRequest
) -> dict:
    return {"endpoint": f"data/v1.1/{id}.csv", "json": request.request_dict()}


def _query_distinct_fields_request_info(
    request: DistinctQueryRequest, use_cache: Optional[bool] = None
) -> dict:
    return {
        "endpoint": f"data/v1.1/distinct",
        "json": request.request_dict(),
        "params": {"cache": str(use_cache)},
    }


def _query_request_info(request: QueryRequest, one: Optional[bool] = False) -> dict:
    return {
        "endpoint": f"data/v1.1/query",
        "json": request.request_dict(),
        "params": {"one": "true"} if one else {},
    }



from typing import Dict, Any, List

from spotlight.api.permission.__util import (
    _get_permission_request_info,
    _delete_permission_request_info,
    _update_permission_request_info,
    _create_permission_request_info,
    _get_permissions_by_resource_id_request_info,
    _get_permission_by_resource_id_request_info,
)
from spotlight.api.permission.model import PermissionRequest
from spotlight.core.common.decorators import data_request
from spotlight.core.common.requests import (
    __get_request,
    __delete_request,
    __post_request,
    __put_request,
)


@data_request()
def get_permission(id: str) -> Dict[str, Any]:
    """
    Get permission by ID.

    Args:
        id (str): Permission ID

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _get_permission_request_info(id)
    return __get_request(**request_info)


@data_request()
def get_permission_by_resource_id(resource_id: str) -> Dict[str, Any]:
    """
    Get permission by resource ID.

    Args:
        resource_id (str): Resource ID

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _get_permission_by_resource_id_request_info(resource_id)
    return __get_request(**request_info)


@data_request()
def get_permissions_by_resource_id(resource_id: str) -> List[Dict[str, Any]]:
    """
    Get all permissions by resource ID.

    Args:
        resource_id (str): Resource ID

    Returns:
        List[Dict[str, Any]]: List of permission response
    """
    request_info = _get_permissions_by_resource_id_request_info(resource_id)
    return __get_request(**request_info)


@data_request()
def create_permission(request: PermissionRequest) -> Dict[str, Any]:
    """
    Create permission.

    Args:
        request (PermissionRequest): Permission request

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _create_permission_request_info(request)
    return __put_request(**request_info)


@data_request()
def update_permission(id: str, request: PermissionRequest) -> Dict[str, Any]:
    """
    Update permission.

    Args:
        id: (str): Permission ID
        request (PermissionRequest): Permission request

    Returns:
        Dict[str, Any]: Permission response
    """
    request_info = _update_permission_request_info(id, request)
    return __post_request(**request_info)


@data_request(processor=lambda response: None)
def delete_permission(id: str) -> None:
    """
    Delete permission by ID.

    Args:
        id: (str): Permission ID

    Returns:
        None
    """
    request_info = _delete_permission_request_info(id)
    return __delete_request(**request_info)



from pydantic import Field

from spotlight.core.common.base import Base
from spotlight.core.common.config import EnvironmentConfig

config = EnvironmentConfig()


class BarchartQuery(Base):
    apikey: str = Field(default=config.auth_config.barchart_api_token, init=False)
    symbol: str
    type: str



"""
Data API query functions for timeseries data.
"""

from spotlight.api.data.asynchronous import (
    async_query_timeseries,
    async_query_timeseries_csv,
    async_query_distinct_fields,
    async_query,
)
from spotlight.api.data.synchronous import (
    query_timeseries,
    query_timeseries_csv,
    query_distinct_fields,
    query,
)



from typing import Optional

from spotlight.api.alert.model import AlertRequest


def _get_alert_request_info(id: str) -> dict:
    return {"endpoint": f"config/alert/{id}"}


def _get_alerts_request_info(override_find_all: Optional[bool] = False) -> dict:
    endpoint = "config/alert"
    params = {"override_find_all": override_find_all}

    filtered_params = {k: v for k, v in params.items() if v is not None}

    return {"endpoint": endpoint, "params": filtered_params}


def _get_alert_signal_request_info(start: int, end: int) -> dict:
    endpoint = "config/alert/signal"
    params = {"start": start, "end": end}

    return {"endpoint": endpoint, "params": params}


def _create_alert_request_info(request: AlertRequest) -> dict:
    return {"endpoint": f"config/alert", "json": request.request_dict()}


def _update_alert_request_info(id: str, request: AlertRequest) -> dict:
    return {"endpoint": f"config/alert/{id}", "json": request.request_dict()}


def _delete_alert_request_info(id: str) -> dict:
    return {"endpoint": f"config/alert/{id}"}



from typing import List, Optional, Any, Dict, Union

import pandas as pd
from requests import Response

from spotlight.api.data.__util import (
    _query_timeseries_csv_request_info,
    _query_timeseries_request_info,
    _query_request_info,
    _query_distinct_fields_request_info,
)
from spotlight.api.data.model import (
    TimeseriesQueryRequest,
    QueryRequest,
    DistinctQueryRequest,
)
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import __async_get_request, __async_post_request


@async_data_request
async def async_query_timeseries(
    request: TimeseriesQueryRequest, one: Optional[bool] = False
) -> List[Dict[str, Any]]:
    """
    Asynchronously get timeseries dataset by timeseries query request.

    Args:
        request (TimeseriesQueryRequest): Timeseries query request
        one (Optional[bool]): Flag to determine whether to return only one record

    Returns:
        pd.DataFrame: Timeseries DataFrame
    """
    request_info = _query_timeseries_request_info(request, one)
    return await __async_post_request(**request_info)


async def async_query_timeseries_csv(
    id: str, request: TimeseriesQueryRequest
) -> Response:
    """
    Asynchronously query dataset CSV by ID.

    Args:
        id (str): Dataset ID
        request (TimeseriesQueryRequest): Timeseries query request

    Returns:
        Response: HTTP response object
    """
    request_info = _query_timeseries_csv_request_info(id, request)
    return await __async_get_request(**request_info)


@async_data_request
async def async_query_distinct_fields(
    request: DistinctQueryRequest, use_cache: Optional[bool] = None
) -> List[str]:
    """
    Asynchronously query dataset for distinct values of a specified field.

    Args:
        request (DistinctQueryRequest): Distinct query request
        use_cache (Optional[bool]): Flag to determine whether data cache should be used in fetching unique values

    Returns:
        pd.DataFrame: Timeseries DataFrame
    """
    request_info = _query_distinct_fields_request_info(request, use_cache)
    return await __async_post_request(**request_info)


@async_data_request
def async_query(
    request: QueryRequest, one: Optional[bool] = False
) -> List[Dict[str, Any]]:
    """
    Asynchronously query dataset by query request.

    Args:
        request (QueryRequest): Query request
        one (Union[bool, str]): Flag to determine whether to return only one record

    Returns:
        List[Dict[str, Any]]: List of records
    """
    request_info = _query_request_info(request, one)
    return __async_post_request(**request_info)



from spotlight.api.data.barchart.model import BarchartQuery


def _transform_results(data: dict) -> dict:
    return data["results"]


def _get_history_info(query: BarchartQuery) -> dict:
    return {
        "endpoint": "/getHistory.json",
        "url_key": "barchart",
        "params": query.request_dict(),
    }



from typing import Optional

from spotlight.api.data_rule.model import DataRuleRequest


def _get_data_rule_request_info(id: str) -> dict:
    return {"endpoint": f"config/data_rule/{id}"}


def _get_data_rules_request_info(override_find_all: Optional[bool] = False) -> dict:
    endpoint = "config/data_rule"
    params = {"override_find_all": override_find_all}

    filtered_params = {k: v for k, v in params.items() if v is not None}

    return {"endpoint": endpoint, "params": filtered_params}


def _create_data_rule_request_info(request: DataRuleRequest) -> dict:
    return {"endpoint": f"config/data_rule", "json": request.request_dict()}


def _update_data_rule_request_info(id: str, request: DataRuleRequest) -> dict:
    return {"endpoint": f"config/data_rule/{id}", "json": request.request_dict()}


def _delete_data_rule_request_info(id: str) -> dict:
    return {"endpoint": f"config/data_rule/{id}"}



"""
Auth API requests.
"""

from typing import Optional

import requests
from requests import Response

from spotlight.api.auth.__util import (
    _login_request_info,
    _refresh_token_request_info,
    _token_exchange_request_info,
)


def login(username: str, password: str, scope: Optional[str] = None) -> Response:
    """
    Login.

    Args:
        username (str): Username
        password (str): Password
        scope (Optional[str]): Token scope, such as 'offline_access'

    Returns:
        dict: Authentication/authorization response
    """
    request_info = _login_request_info(username, password, scope)
    return requests.post(**request_info)


def refresh_token(token: Optional[str]) -> Response:
    """
    Refresh token. To get an offline access token, initial authentication (login) needs to use the scope
    'offline_access', then an offline token can be created using the refresh_token in the authentication response.
    [Stack overflow reference](https://stackoverflow.com/questions/69207734/keycloak-offline-access-token-with-refresh-token-grant-type).

    Args:
        token (str): Refresh token

    Returns:
        dict: Authentication/authorization response
    """
    request_info = _refresh_token_request_info(token)
    return requests.post(**request_info)


def exchange_token(requested_subject: str, subject_token: str) -> Response:
    """
    Exchange token.

    Args:
        requested_subject (str): Request subject (Username or ID)
        subject_token (str): Token to swap

    Returns:
        dict: Authentication/authorization response
    """
    request_info = _token_exchange_request_info(requested_subject, subject_token)
    return requests.post(**request_info)



"""
Data classes for API requests.
"""

from typing import List, Optional, Any

from pydantic import Field as PyField

from spotlight.core.common.base import Base
from spotlight.core.common.decorators.serializable import (
    serializable_base_class,
    serializable,
)
from spotlight.core.common.enum import (
    Order,
    ComparisonOperator,
    LogicalOperator,
    SqlFunction,
)


class Sort(Base):
    field: str
    order: Order


class Filter(Base):
    field: str
    operator: ComparisonOperator
    value: Any


class WhereClause(Base):
    filter: Optional[Filter] = PyField(default=None)
    operator: Optional[LogicalOperator] = PyField(default=None)
    left: Optional["WhereClause"] = PyField(default=None)
    right: Optional["WhereClause"] = PyField(default=None)


class TimeseriesQueryRequest(Base):
    id: Optional[str] = PyField(default=None)
    dataset_name: Optional[str] = PyField(default=None)
    reference_name: Optional[str] = PyField(default=None)
    page: Optional[int] = PyField(default=None)
    limit: Optional[int] = PyField(default=None)
    fields: Optional[List[str]] = PyField(default=None)
    sort: Optional[List[Sort]] = PyField(default=None)
    where: Optional[WhereClause] = PyField(default=None)


class DistinctQueryRequest(Base):
    id: Optional[str] = PyField(default=None)
    dataset_name: Optional[str] = PyField(default=None)
    reference_name: Optional[str] = PyField(default=None)
    field: str = PyField(default=None)
    sort: Optional[List[Sort]] = PyField(default=None)
    where: Optional[WhereClause] = PyField(default=None)


@serializable_base_class
class Expression(Base):
    pass


@serializable
class FieldExpression(Expression):
    pass


@serializable
class FunctionExpression(FieldExpression):
    pass


@serializable
class Field(FieldExpression):
    name: str
    alias: Optional[str] = PyField(default=None)


@serializable
class SingleExpression(FunctionExpression):
    parameter: FieldExpression
    operator: SqlFunction
    alias: Optional[str] = PyField(default=None)


@serializable
class MultiExpression(FunctionExpression):
    parameters: List[FieldExpression]
    operator: SqlFunction
    alias: Optional[str] = PyField(default=None)


class QueryRequest(Base):
    id: Optional[str] = PyField(default=None)
    dataset_name: Optional[str] = PyField(default=None)
    reference_name: Optional[str] = PyField(default=None)
    fields: Optional[List[FieldExpression]] = PyField(default=None)
    where: Optional[WhereClause] = PyField(default=None)
    groups: Optional[List[str]] = PyField(default=None)
    page: Optional[int] = PyField(default=None)
    limit: Optional[int] = PyField(default=None)
    sort: Optional[List[Sort]] = PyField(default=None)



from spotlight.api.permission.model import PermissionRequest


def _get_permission_request_info(id: str) -> dict:
    return {"endpoint": f"config/permission/{id}"}


def _get_permission_by_resource_id_request_info(resource_id: str) -> dict:
    return {"endpoint": f"config/permission/resource/{resource_id}"}


def _get_permissions_by_resource_id_request_info(resource_id: str) -> dict:
    return {"endpoint": f"config/permission/resource/{resource_id}/permissions"}


def _create_permission_request_info(request: PermissionRequest) -> dict:
    return {"endpoint": f"config/permission", "json": request.request_dict()}


def _update_permission_request_info(id: str, request: PermissionRequest) -> dict:
    return {"endpoint": f"config/permission/{id}", "json": request.request_dict()}


def _delete_permission_request_info(id: str) -> dict:
    return {"endpoint": f"config/permission/{id}"}



from spotlight.api.dataset.model import DatasetRequest
from spotlight.api.dataset.model import SearchRequest


def _get_dataset_request_info(id: str) -> dict:
    return {"endpoint": f"config/dataset/{id}"}


def _get_datasets_request_info() -> dict:
    return {"endpoint": f"config/dataset"}


def _search_datasets_request_info(request: SearchRequest) -> dict:
    return {
        "endpoint": f"config/dataset/search",
        "json": request.request_dict(),
    }


def _create_dataset_request_info(request: DatasetRequest) -> dict:
    return {"endpoint": f"config/dataset", "json": request.request_dict()}


def _update_dataset_request_info(id: str, request: DatasetRequest) -> dict:
    return {"endpoint": f"config/dataset/{id}", "json": request.request_dict()}


def _delete_dataset_request_info(id: str) -> dict:
    return {"endpoint": f"config/dataset/{id}"}



"""
Common enums.
"""

from spotlight.core.common.base_enum import BaseEnum


class Order(BaseEnum):
    ASC = "ASC"
    DESC = "DESC"


class ComparisonOperator(BaseEnum):
    NOT_IN = "NOT_IN"
    IN = "IN"
    NOT_EQUAL = "NOT_EQUAL"
    EQUAL = "EQUAL"
    LIKE = "LIKE"
    NOT_LIKE = "NOT_LIKE"
    GTE = "GTE"
    GT = "GT"
    LTE = "LTE"
    LT = "LT"


class LogicalOperator(BaseEnum):
    AND = "AND"
    OR = "OR"
    NOT = "NOT"


class SqlFunction(BaseEnum):
    SUM = "SUM"
    AVG = "AVG"
    ABS = "ABS"
    AND = "AND"
    COUNT = "COUNT"
    COUNT_DISTINCT = "COUNT_DISTINCT"
    MIN = "MIN"
    MAX = "MAX"
    MEDIAN = "MEDIAN"
    COALESCE = "COALESCE"


class FieldType(BaseEnum):
    BOOLEAN = "BOOLEAN"
    DOUBLE = "DOUBLE"
    INTEGER = "INTEGER"
    TEXT = "TEXT"
    TIMESTAMP = "TIMESTAMP"
    DATETIME = "DATETIME"


class UpdateRequestAction(BaseEnum):
    UPDATE = "UPDATE"
    COPY = "COPY"


class RuleType(BaseEnum):
    SQL = "SQL"
    SPARK = "SPARK"
    WHERE_CLAUSE = "WHERE_CLAUSE"


class RuleSeverity(BaseEnum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"


class EventType(BaseEnum):
    CREATE = "CREATE"
    UPDATE = "UPDATE"
    DELETE = "DELETE"


class IntervalType(BaseEnum):
    DAY = "DAY"
    HOUR = "HOUR"
    MINUTE = "MINUTE"


class NotificationType(BaseEnum):
    EMAIL = "EMAIL"


class Repository(BaseEnum):
    CME = "CME"
    DTCC = "DTCC"
    ICE = "ICE"


class AssetClass(BaseEnum):
    COMMODITY = "COMMODITY"
    CREDIT = "CREDIT"
    EQUITY = "EQUITY"
    FOREX = "FOREX"
    RATES = "RATES"



from typing import Optional, List

from pydantic import Field as pyField

from spotlight.api.data.model import WhereClause, Sort
from spotlight.core.common.base import Base
from spotlight.core.common.enum import FieldType


class Field(Base):
    display_name: str
    logical_name: str
    type: FieldType
    width: int
    editable: bool
    hidden: bool
    display_order: int
    tags: List[str]
    field_group_display_name: str
    field_group_logical_name: str


class DatasetRequest(Base):
    display_name: str
    description: Optional[str]
    filter_field: Optional[str]
    reference_name: Optional[str]
    tags: Optional[List[str]]
    where_clause: Optional[WhereClause]
    row_limit: Optional[int]
    max_limit: Optional[int]
    sort: Optional[List[Sort]] = pyField(default=None)
    schema_: Optional[List[Field]] = pyField(alias="schema")
    dataset_id: str
    custom: bool


class DatasetResponse(Base):
    id: str
    display_name: str
    description: Optional[str]
    filter_field: Optional[str]
    reference_name: Optional[str]
    tags: List[str]
    schema_: Optional[List[Field]] = pyField(alias="schema")
    where_clause: Optional[WhereClause]
    row_limit: Optional[int]
    max_limit: Optional[int]
    sort: Optional[List[Sort]] = pyField(default=None)
    abstract_dataset_id: str
    custom: bool
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]


class SearchRequest(Base):
    query: str



"""
Auth API.
"""

from spotlight.api.auth.asynchronous import (
    async_login,
    async_refresh_token,
    async_exchange_token,
)
from spotlight.api.auth.synchronous import login, refresh_token, exchange_token



"""
Base classes with custom attributes for updating, serializing and deserializing data classes and enums.
"""

import base64
import copy
import json
from abc import ABC
from datetime import date, datetime, tzinfo
from typing import List

import pydash as _
from pydantic import BaseModel
from pydantic.utils import deep_update

from spotlight.core.common.date import date_to_timestamp, datetime_to_timestamp
from spotlight.core.common.function import to_nested_dict


class Base(ABC, BaseModel):
    """
    Base class used for all data classes.
    """

    class Config:
        # use enum values when using .dict() on object
        use_enum_values = True

        json_encoders = {
            date: date_to_timestamp,
            datetime: datetime_to_timestamp,
            tzinfo: str,
        }

    @classmethod
    def cls_name(cls) -> str:
        """
        Get class name.

        Returns:
            str: Class name
        """
        return cls.__name__

    def request_dict(self) -> dict:
        """
        Convert data class to dict. Used instead of `.dict()` to serialize dates as timestamps.

        Returns:
            dict: Serialized data class as dict
        """
        return json.loads(self.json(by_alias=True))

    def base64_encoded(self, exclude=None) -> bytes:
        """
        Base-64 encode data class.

        Returns:
            bytes: Base-64 encoded data class as bytes
        """
        json_str = json.dumps(self.json(exclude=exclude), sort_keys=True)
        bytes_rep = bytes(json_str, "utf-8")
        return base64.b64encode(bytes_rep)

    def __hash__(self):
        """
        Pydantic doesn't support hashing its base models so this is a work around

        https://stackoverflow.com/a/63774573/8189527
        """
        return hash((type(self),) + tuple(self.__dict__.items()))

    def copy(self, ignored_fields: List[str] = None, **kwargs) -> "Base":
        """
        Create a copy of the object

        Parameters:
            ignored_fields: fields to ignore from the original object when copying
            **kwargs: key value pairs of fields you want to change during the copy, for nested fields delimit the keys
            with a period (e.g. `.`)

        Returns:
            Copy of the object
        """
        cls = self.__class__
        obj_dict = copy.deepcopy(self.request_dict())
        ignored_fields = [] if ignored_fields is None else ignored_fields

        updated_fields = to_nested_dict(kwargs)
        obj_dict = deep_update(obj_dict, updated_fields)
        [_.unset(obj_dict, field) for field in ignored_fields]
        return cls(**obj_dict)



import abc


class Singleton(type):
    """
    This is a metaclass used for turning classes into singleton objects

    ref: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

    Usage:
        class FooSingleton(metaclass=Singleton):
            def __init__(self):
                self.x = 5

        foo = FooSingleton()
        foo2 = FooSingleton()

        assert foo is foo2
        assert foo.x == foo2.x
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class ABCSingleton(Singleton, abc.ABCMeta):
    """
    This singleton object is used when the class or the classes parent utilizes ABC

    ref: https://stackoverflow.com/questions/57349105/python-abc-inheritance-with-specified-metaclass
    """

    pass



from spotlight.api.dataset.view.asynchronous import (
    async_get_dataset_view,
    async_get_dataset_views,
    async_search_dataset_views,
)
from spotlight.api.dataset.view.synchronous import (
    get_dataset_view,
    get_dataset_views,
    search_dataset_views,
)



from spotlight.api.dataset.asynchronous import (
    async_get_dataset,
    async_get_datasets,
    async_create_dataset,
    async_update_dataset,
    async_delete_dataset,
    async_search_datasets,
)
from spotlight.api.dataset.synchronous import (
    get_dataset,
    get_datasets,
    create_dataset,
    update_dataset,
    delete_dataset,
    search_datasets,
)



"""
Configuration classes used for configuring the Spotlight SDK.
"""

import os
from typing import Optional


class Auth:
    """
    Configuration class for getting the authentication information from the environment.
    """

    # Feature to keep values hidden from printing (not totally)
    # Allows for dynamic rendering of auth values
    @property
    def username(self) -> Optional[str]:
        return os.getenv("SPOTLIGHT_USER", None)

    @property
    def password(self) -> Optional[str]:
        return os.getenv("SPOTLIGHT_PWD", None)

    @property
    def access_token(self) -> Optional[str]:
        return os.getenv("SPOTLIGHT_ACCESS_TOKEN", None)

    @property
    def refresh_token(self) -> Optional[str]:
        return os.getenv("SPOTLIGHT_REFRESH_TOKEN", None)

    @property
    def barchart_api_token(self) -> Optional[str]:
        return os.getenv("SPOTLIGHT_BARCHART_API_TOKEN", None)

    def can_authenticate(self) -> bool:
        """
        Check environment to see if the user provided enough info to authenticate requests.

        Returns:
            bool: A boolean that represents whether you can authenticate requests
        """
        auth_options = (
            self.access_token or self.refresh_token or (self.username and self.password)
        )
        return auth_options is not None

    def __eq__(self, other: "Auth"):
        return (
            self.access_token == other.access_token
            and self.username == self.username
            and self.password == self.password
        )


class EnvironmentConfig:
    """
    Configuration class for getting environment information.
    """

    auth_config: Auth = Auth()

    @property
    def environment(self) -> str:
        env = os.getenv("SPOTLIGHT_ENV", "production")
        return env

    @property
    def urls(self) -> dict:
        env_url_map = {
            "production": {
                "spotlight": "https://api.spotlight.dev",
                "keycloak_admin": "https://auth.spotlight.dev/auth/admin/realms/production",
                "keycloak": "https://auth.spotlight.dev/auth/realms/production",
                "barchart": "https://ondemand.websol.barchart.com",
            },
            "development": {
                "spotlight": "https://api.dev.spotlight.dev",
                "keycloak_admin": "https://auth.dev.spotlight.dev/auth/admin/realms/production",
                "keycloak": "https://auth.dev.spotlight.dev/auth/realms/production",
                "barchart": "https://ondemand.websol.barchart.com",
            },
        }
        return env_url_map.get(self.environment, env_url_map["production"])

    def get_url(self, key: str = "spotlight") -> str:
        """
        Get base URL given key.

        Args:
            key (str): URL key; 'spotlight', 'keycloak_admin', 'keycloak' or 'fred'

        Returns:
            str: Base URL string
        """
        return self.urls.get(key, self.urls["spotlight"])



from spotlight.api.dataset.abstract.model import AbstractDatasetRequest
from spotlight.api.dataset.model import SearchRequest


def _get_abstract_dataset_request_info(id: str) -> dict:
    return {"endpoint": f"config/dataset/abstract/{id}"}


def _get_abstract_datasets_request_info() -> dict:
    return {"endpoint": f"config/dataset/abstract"}


def _search_abstract_datasets_request_info(request: SearchRequest) -> dict:
    return {
        "endpoint": f"config/dataset/abstract/search",
        "json": request.request_dict(),
    }


def _create_abstract_dataset_request_info(request: AbstractDatasetRequest) -> dict:
    return {"endpoint": f"config/dataset/abstract", "json": request.request_dict()}


def _update_abstract_dataset_request_info(
    id: str, request: AbstractDatasetRequest
) -> dict:
    return {
        "endpoint": f"config/dataset/abstract/{id}",
        "json": request.request_dict(),
    }


def _delete_abstract_dataset_request_info(id: str) -> dict:
    return {"endpoint": f"config/dataset/abstract/{id}"}



from typing import Dict, Any, List

from spotlight.api.dataset.abstract.__util import (
    _get_abstract_dataset_request_info,
    _get_abstract_datasets_request_info,
    _search_abstract_datasets_request_info,
    _create_abstract_dataset_request_info,
    _update_abstract_dataset_request_info,
    _delete_abstract_dataset_request_info,
)
from spotlight.api.dataset.abstract.model import AbstractDatasetRequest
from spotlight.api.dataset.model import SearchRequest
from spotlight.core.common.decorators import data_request
from spotlight.core.common.requests import (
    __get_request,
    __post_request,
    __put_request,
    __delete_request,
)


@data_request()
def get_abstract_dataset(id: str) -> Dict[str, Any]:
    """
    Get abstract dataset by ID.

    Args:
        id (str): Abstract dataset ID

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _get_abstract_dataset_request_info(id)
    return __get_request(**request_info)


@data_request()
def get_abstract_datasets() -> List[Dict[str, Any]]:
    """
    Get all abstract datasets.

    Returns:
        List[Dict[str, Any]]: List of abstract dataset response
    """
    request_info = _get_abstract_datasets_request_info()
    return __get_request(**request_info)


@data_request()
def search_abstract_datasets(request: SearchRequest) -> List[Dict[str, Any]]:
    """
    Search all abstract datasets.

    Args:
        request (SearchRequest): Abstract dataset search request

    Returns:
        List[Dict[str, Any]]: List of abstract dataset response
    """
    request_info = _search_abstract_datasets_request_info(request)
    return __post_request(**request_info)


@data_request()
def create_abstract_dataset(request: AbstractDatasetRequest) -> Dict[str, Any]:
    """
    Create abstract dataset.

    Args:
        request (DatasetRequest): Abstract dataset request

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _create_abstract_dataset_request_info(request)
    return __put_request(**request_info)


@data_request()
def update_abstract_dataset(id: str, request: AbstractDatasetRequest) -> Dict[str, Any]:
    """
    Update abstract dataset.

    Args:
        id (str): Abstract dataset ID
        request (DatasetRequest): Abstract dataset request

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _update_abstract_dataset_request_info(id, request)
    return __post_request(**request_info)


@data_request(processor=lambda response: None)
def delete_abstract_dataset(id: str) -> None:
    """
    Delete abstract dataset by ID.

    Args:
        id (str): Abstract dataset ID

    Returns:
        None
    """
    request_info = _delete_abstract_dataset_request_info(id)
    return __delete_request(**request_info)



from typing import Dict, Any, List

from spotlight.api.dataset.model import SearchRequest
from spotlight.api.dataset.view.__util import (
    _get_dataset_view_request_info,
    _get_dataset_views_request_info,
    _search_dataset_views_request_info,
)
from spotlight.core.common.decorators import data_request
from spotlight.core.common.requests import (
    __get_request,
    __post_request,
)


@data_request()
def get_dataset_view(id: str) -> Dict[str, Any]:
    """
    Get dataset view by ID.

    Args:
        id (str): Dataset ID

    Returns:
        Dict[str, Any]: Dataset view response
    """
    request_info = _get_dataset_view_request_info(id)
    return __get_request(**request_info)


@data_request()
def get_dataset_views() -> List[Dict[str, Any]]:
    """
    Get all dataset views.

    Returns:
        List[Dict[str, Any]]: List of dataset view response
    """
    request_info = _get_dataset_views_request_info()
    return __get_request(**request_info)


@data_request()
def search_dataset_views(request: SearchRequest) -> List[Dict[str, Any]]:
    """
    Search all dataset views.

    Args:
        request (SearchRequest): Dataset view search request

    Returns:
        List[Dict[str, Any]]: List of dataset view response
    """
    request_info = _search_dataset_views_request_info(request)
    return __post_request(**request_info)



from spotlight.api.dataset.model import SearchRequest


def _get_dataset_view_request_info(id: str) -> dict:
    return {"endpoint": f"config/dataset/view/{id}"}


def _get_dataset_views_request_info() -> dict:
    return {"endpoint": f"config/dataset/view"}


def _search_dataset_views_request_info(request: SearchRequest) -> dict:
    return {
        "endpoint": f"config/dataset/view/search",
        "json": request.request_dict(),
    }



from typing import Optional

import requests
from requests import Response

from spotlight.core.common.config import EnvironmentConfig
from spotlight.core.common.decorators import authenticated_request

config = EnvironmentConfig()


@authenticated_request
def __get_request(
    endpoint: str,
    url_key: str = "spotlight",
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for get requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    return requests.get(
        url=f"{config.get_url(url_key)}/{endpoint}",
        headers=headers or {},
        params=params,
        timeout=timeout,
    )


@authenticated_request
def __put_request(
    endpoint: str,
    url_key: str = "spotlight",
    data=None,
    json=None,
    files=None,
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for put requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    return requests.put(
        url=f"{config.get_url(url_key)}/{endpoint}",
        data=data,
        json=json,
        files=files,
        headers=headers or {},
        params=params,
        timeout=timeout,
    )


@authenticated_request
def __post_request(
    endpoint: str,
    url_key: str = "spotlight",
    files=None,
    data=None,
    json=None,
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for post requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    return requests.post(
        url=f"{config.get_url(url_key)}/{endpoint}",
        files=files,
        data=data,
        json=json,
        headers=headers or {},
        params=params,
        timeout=timeout,
    )


@authenticated_request
def __delete_request(
    endpoint: str,
    url_key: str = "spotlight",
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for delete requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    return requests.delete(
        url=f"{config.get_url(url_key)}/{endpoint}",
        headers=headers or {},
        params=params,
        timeout=timeout,
    )



import functools
from typing import Tuple, Dict, Any, Callable, List

from requests import Response

from spotlight.core.common.decorators.data.__util import __process_response
from spotlight.core.common.function import no_transform


def async_data_request(
    request: Callable = None,
    *,
    processor: Callable[[Response], Any] = lambda response: response.json(),
    transformers: List[Callable[[Any], Any]] = None
):
    """
    Decorator for processing the response of an asynchronous REST request.

    Args:
        request (Callable): Function
        processor (Callable[[Response], Any]): Transform response
        transformers (Callable[[Any], Any]): Transform data output

    Returns:
        Callable: Wrapped function
    """

    @functools.wraps(request)
    def decorator(f):
        @functools.wraps(f)
        async def wrap(*args: Tuple, **kwargs: Dict[str, Any]) -> Any:
            _processor = kwargs.pop("processor", None) or processor
            _transformers = (
                kwargs.pop("transformers", None)
                or transformers
                or [lambda data: no_transform(data)]
            )
            response: Response = await f(*args, **kwargs)
            return __process_response(response, _processor, _transformers)

        return wrap

    if request is None:
        return decorator
    return decorator(request)



"""
Custom exceptions.
"""


class CustomBaseException(Exception):
    """
    Base exception for all sub-exceptions.
    """

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)

    def __str__(self):
        return self.message


class AuthenticationError(CustomBaseException):
    """
    Custom exception for authentication errors when interacting with the APIs.
    """

    def __init__(self, message: str, *args, **kwargs):
        super().__init__(message)


class ApiError(CustomBaseException):
    """
    Custom exception for non-authentication related API errors.
    """

    def __init__(self, message: str, *args, **kwargs):
        super().__init__(message)



from typing import Dict, Any, List

from spotlight.api.dataset.__util import (
    _get_dataset_request_info,
    _get_datasets_request_info,
    _search_datasets_request_info,
    _create_dataset_request_info,
    _update_dataset_request_info,
    _delete_dataset_request_info,
)
from spotlight.api.dataset.model import DatasetRequest
from spotlight.api.dataset.model import SearchRequest
from spotlight.core.common.decorators import data_request
from spotlight.core.common.requests import (
    __get_request,
    __post_request,
    __put_request,
    __delete_request,
)


@data_request()
def get_dataset(id: str) -> Dict[str, Any]:
    """
    Get dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _get_dataset_request_info(id)
    return __get_request(**request_info)


@data_request()
def get_datasets() -> List[Dict[str, Any]]:
    """
    Get all dataset.

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _get_datasets_request_info()
    return __get_request(**request_info)


@data_request()
def search_datasets(request: SearchRequest) -> List[Dict[str, Any]]:
    """
    Search all dataset.

    Args:
        request (SearchRequest): Dataset search request

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _search_datasets_request_info(request)
    return __post_request(**request_info)


@data_request()
def create_dataset(request: DatasetRequest) -> Dict[str, Any]:
    """
    Create dataset.

    Args:
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _create_dataset_request_info(request)
    return __put_request(**request_info)


@data_request()
def update_dataset(id: str, request: DatasetRequest) -> Dict[str, Any]:
    """
    Update dataset.

    Args:
        id (str): Dataset ID
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _update_dataset_request_info(id, request)
    return __post_request(**request_info)


@data_request(processor=lambda response: None)
def delete_dataset(id: str) -> None:
    """
    Delete dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        None
    """
    request_info = _delete_dataset_request_info(id)
    return __delete_request(**request_info)



from spotlight.api.dataset.abstract.asynchronous import (
    async_get_abstract_dataset,
    async_get_abstract_datasets,
    async_search_abstract_datasets,
    async_create_abstract_dataset,
    async_update_abstract_dataset,
    async_delete_abstract_dataset,
)
from spotlight.api.dataset.abstract.synchronous import (
    get_abstract_dataset,
    get_abstract_datasets,
    search_abstract_datasets,
    create_abstract_dataset,
    update_abstract_dataset,
    delete_abstract_dataset,
)



from datetime import datetime
from typing import Optional, Union

from pydantic import Field

from spotlight.api.data.model import WhereClause
from spotlight.core.common.base import Base
from spotlight.core.common.enum import RuleSeverity, RuleType, EventType


class DataRuleRequest(Base):
    display_name: str
    dataset_id: str
    severity: RuleSeverity
    type: RuleType
    predicate: Optional[Union[str, WhereClause]] = Field(default=None)


class DataRuleResponse(Base):
    id: str
    display_name: str
    dataset_id: str
    severity: RuleSeverity
    type: RuleType
    predicate: Optional[Union[str, WhereClause]] = Field(default=None)
    created_by: str
    created_at: datetime
    updated_by: Optional[str] = Field(default=None)
    updated_at: Optional[int] = Field(default=None)


class DataRuleEvent(Base):
    rule: DataRuleResponse
    type: EventType



from typing import Optional

import aiohttp
from requests import Response

from spotlight.core.common.config import EnvironmentConfig
from spotlight.core.common.decorators import async_authenticated_request
from spotlight.core.common.function import build_response_obj

config = EnvironmentConfig()


@async_authenticated_request
async def __async_get_request(
    endpoint: str,
    url_key: str = "spotlight",
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for asynchronous get requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    async with aiohttp.ClientSession() as session:
        async with session.get(
            url=f"{config.get_url(url_key)}/{endpoint}",
            headers=headers or {},
            params=params,
            timeout=timeout,
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_put_request(
    endpoint: str,
    url_key: str = "spotlight",
    data=None,
    json=None,
    files=None,
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for asynchronous put requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    if files:
        data = files

    async with aiohttp.ClientSession() as session:
        async with session.put(
            url=f"{config.get_url(url_key)}/{endpoint}",
            data=data,
            json=json,
            headers=headers or {},
            params=params,
            timeout=timeout,
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_post_request(
    endpoint: str,
    url_key: str = "spotlight",
    files=None,
    data=None,
    json=None,
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for asynchronous post requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        data (Any): Data passed to the API
        json (Any): JSON passed to the API
        files (Any): Files passed to the API
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    if files:
        data = files

    async with aiohttp.ClientSession() as session:
        async with session.post(
            url=f"{config.get_url(url_key)}/{endpoint}",
            data=data,
            json=json,
            headers=headers or {},
            params=params,
            timeout=timeout,
        ) as response:
            result = await build_response_obj(response)

    return result


@async_authenticated_request
async def __async_delete_request(
    endpoint: str,
    url_key: str = "spotlight",
    headers: Optional[dict] = None,
    params: Optional[dict] = None,
    timeout: int = 120,
) -> Response:
    """
    Wrapper method for asynchronous delete requests.

    Args:
        endpoint (str): API endpoint
        url_key (str): Key for looking up the url from the environment config
        headers (Optional[dict]): Headers for the API request
        params (Optional[dict]): Params for the API request
        timeout (int): Timeout for API request

    Returns:
        Response: HTTP response object
    """
    async with aiohttp.ClientSession() as session:
        async with session.delete(
            url=f"{config.get_url(url_key)}/{endpoint}",
            headers=headers or {},
            params=params,
            timeout=timeout,
        ) as response:
            result = await build_response_obj(response)

    return result



from pandas import DataFrame

from spotlight.api.data.barchart.__util import _get_history_info, _transform_results
from spotlight.api.data.barchart.model import BarchartQuery
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import __async_get_request


@async_data_request(transformers=[_transform_results])
async def async_get_history(query: BarchartQuery):
    """
    Asynchronously get barchart history.

    Args:
        query (BarchartQuery): Barchart query

    Returns:
        DataFrame: DataFrame
    """
    request_info = _get_history_info(query)
    return await __async_get_request(**request_info)



from spotlight.api.data_rule.asynchronous import (
    async_get_data_rule,
    async_get_data_rules,
    async_create_data_rule,
    async_update_data_rule,
    async_delete_data_rule,
)
from spotlight.api.data_rule.synchronous import (
    get_data_rule,
    get_data_rules,
    create_data_rule,
    update_data_rule,
    delete_data_rule,
)



from datetime import date
from typing import Dict, Any, List

from spotlight.api.sdr_source.__util import (
    _sdr_metadata_request_info,
    _sdr_upload_request_info,
    _sdr_download_request_info,
    _sdr_download_json_request_info,
    _sdr_cache_clear_file_request_info,
    _sdr_cache_clear_record_request_info,
    _sdr_cache_lookup_request_info,
    _sdr_status_request_info,
)
from spotlight.core.common.decorators import data_request
from spotlight.core.common.enum import AssetClass
from spotlight.core.common.requests import (
    __get_request,
    __post_request,
)


@data_request()
def get_metadata(
    asset_class: AssetClass, from_date: date, to_date: date, reprocess: bool = True
) -> List[Dict[str, Any]]:
    """
    Get metadata between two dates.

    Args:
        asset_class (AssetClass): Asset class
        from_date (date): Start date
        to_date (date): End date
        reprocess (bool): Reprocess flag

    Returns:
        List[Dict[str, Any]]: Metadata response
    """
    request_info = _sdr_metadata_request_info(
        asset_class, from_date, to_date, reprocess
    )
    return __get_request(**request_info)


@data_request()
def upload_metadata(metadata: List[Dict[str, Any]]) -> None:
    """
    Upload sdr data for metadata.

    Args:
        metadata (List[Dict[str, Any]]): Metadata

    Returns:
        None
    """
    request_info = _sdr_upload_request_info(metadata)
    return __post_request(**request_info)


@data_request(processor=lambda r: r.text)
def download_csv(metadata: Dict[str, Any]) -> str:
    """
    Download CSV sdr data for metadata.

    Args:
        metadata (Dict[str, Any]): Metadata

    Returns:
        str: CSV data
    """
    request_info = _sdr_download_request_info(metadata)
    return __post_request(**request_info)


@data_request()
def download_json(metadata: Dict[str, Any]) -> List[Dict[str, Any]]:
    """
    Download JSON sdr data for metadata.

    Args:
        metadata (Dict[str, Any]): Metadata

    Returns:
        List[Dict[str, Any]]: JSON data
    """
    request_info = _sdr_download_json_request_info(metadata)
    return __post_request(**request_info)


@data_request()
def clear_file_cache() -> None:
    """
    Clear file cache.

    Returns:
        None
    """
    request_info = _sdr_cache_clear_file_request_info()
    return __post_request(**request_info)


@data_request()
def clear_record_cache() -> None:
    """
    Clear record cache.

    Returns:
        None
    """
    request_info = _sdr_cache_clear_record_request_info()
    return __post_request(**request_info)


@data_request()
def lookup_cache(metadata: Dict[str, Any]) -> bool:
    """
    Lookup cache for metadata.

    Args:
        metadata (Dict[str, Any]): Metadata

    Returns:
        bool: Exists in cache
    """
    request_info = _sdr_cache_lookup_request_info(metadata)
    return __post_request(**request_info)


@data_request()
def get_status() -> Dict[str, Any]:
    """
    Get scraper status.

    Returns:
        Dict[str, Any]: Health status
    """
    request_info = _sdr_status_request_info()
    return __get_request(**request_info)



from spotlight.api.alert.asynchronous import (
    async_get_alert,
    async_get_alerts,
    async_get_alert_signal,
    async_create_alert,
    async_update_alert,
    async_delete_alert,
)
from spotlight.api.alert.synchronous import (
    get_alert,
    get_alerts,
    get_alert_signal,
    create_alert,
    update_alert,
    delete_alert,
)



import json
import logging

from requests import Response

from spotlight.core.common.errors import AuthenticationError

logger = logging.getLogger(__name__)


def _handle_auth_response(auth_response: Response, error_msg: str) -> dict:
    if auth_response.status_code == 200:
        bearer_token = auth_response.json()["access_token"]
        return {"Authorization": f"Bearer {bearer_token}"}
    else:
        error = AuthenticationError(
            f"Status Code: {auth_response.status_code}\nResponse content: {auth_response.content}\n{error_msg}"
        )
        logger.error(error)
        raise error


def _access_token(access_token: str) -> dict:
    return json.loads(access_token)


def _validate_headers(**kwargs):
    headers = kwargs.get("headers", {})

    if headers.get("Authorization") is None:
        error = AuthenticationError(
            "Authentication for the package was configured incorrectly and is either "
            "missing a SPOTLIGHT_ACCESS_TOKEN or SPOTLIGHT_REFRESH_TOKEN or SPOTLIGHT_USER and "
            "SPOTLIGHT_PWD environment variable(s)"
        )
        logger.error(error)
        # TODO this fails when using FRED API (outside of Spotlight API, so it doesn't need Authorization header)
        #  raise error



"""
Date and time functions.
"""

import calendar
from datetime import date, datetime, time

import pytz

from spotlight.core.common.type import Datetime


def current_timestamp() -> int:
    """ """
    now = datetime.now()
    return datetime_to_timestamp(now)


def date_to_timestamp(as_of_date: date) -> int:
    """
    Convert date to epoch timestamp in milliseconds.

    Args:
        as_of_date (date): Python date

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(as_of_date.timetuple()) * 1000


def datetime_to_timestamp(dt: datetime) -> int:
    """
    Convert datetime to epoch timestamp in milliseconds.

    Args:
        dt (datetime): Python datetime

    Returns:
        int: Epoch timestamp in milliseconds
    """
    return calendar.timegm(dt.utctimetuple()) * 1000


def datetime_to_utc(dt: datetime) -> datetime:
    """
    Standardize datetime to UTC. Assume that datetime where `tzinfo=None` is already in UTC.

    Args:
        dt (datetime): Python datetime

    Returns:
        datetime: Python datetime with standardized UTC timezone (`tzinfo=None`)
    """
    # assume that datetime without timezone is already in UTC
    if dt.tzinfo is None:
        return dt
    return dt.astimezone(pytz.utc).replace(tzinfo=None)


def date_to_datetime(
    as_of_date: date, as_of_time: time = datetime.min.time()
) -> datetime:
    """
    Convert date and optional time to datetime. NOTE: time should not contain a timezone or else offset may not be
    correct.

    Args:
        as_of_date (date): Python date
        as_of_time (time): Python time

    Returns:
        datetime: Python datetime
    """
    return datetime.combine(as_of_date, as_of_time)


def standardize_date_or_datetime(as_of_date: Datetime) -> datetime:
    """
    Convert Datetime (Union[date, datetime]) to standardized UTC datetime.

    Args:
        as_of_date (date): Datetime

    Returns:
        datetime: Python UTC datetime
    """
    if isinstance(as_of_date, datetime):
        return datetime_to_utc(as_of_date)

    return date_to_datetime(as_of_date)


def timestamp_to_datetime(timestamp: int) -> datetime:
    """
    Transform epoch timestamp in milliseconds to datetime.

    Args:
        timestamp (int): Epoch timestamp in milliseconds

    Returns:
        datetime: Python datetime
    """
    return datetime.fromtimestamp(timestamp / 1000)



from typing import Optional, List

from pydantic import Field as pyField

from spotlight.api.data.model import WhereClause, Sort
from spotlight.api.dataset.model import Field
from spotlight.core.common.base import Base


class AbstractDatasetRequest(Base):
    dataset_name: str
    display_name: str
    description: Optional[str]
    table_name: str
    where_clause: Optional[WhereClause]
    row_limit: int
    max_limit: int
    sort: Optional[List[Sort]] = pyField(default=None)
    tags: List[str]
    schema_: List[Field] = pyField(alias="schema")


class AbstractDatasetResponse(Base):
    id: str
    dataset_name: str
    display_name: str
    description: Optional[str]
    table_name: str
    where_clause: Optional[WhereClause]
    row_limit: int
    max_limit: int
    sort: Optional[List[Sort]] = pyField(default=None)
    tags: List[str]
    created_by: str
    created_at: int
    updated_by: Optional[str]
    updated_at: Optional[int]
    schema_: List[Field] = pyField(alias="schema")



import subprocess

import setuptools

try:
    import pypandoc

    long_description = pypandoc.convert_file('README.md', 'rst')
except(IOError, ImportError):
    long_description = open('README.md').read()


def get_tag():
    tag = subprocess.getoutput('git tag --sort=version:refname | tail -n1')
    commits = subprocess.getoutput(f'git rev-list {tag}..HEAD --count')
    return f'{tag}.{commits}'


setuptools.setup(
    name="spotlight-sdk",
    version=get_tag(),
    author="Spotlight",
    author_email="hello@spotlight.dev",
    description="Spotlight Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://spotlight.dev",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.7",
    packages=setuptools.find_packages(
        include=['spotlight*'],
        exclude=['tests.*']
    ),
    install_requires=[
        "requests>=2.31.0",
        "aiohttp>=3.8.4",
        "pandas>=2.0.1",
        "duckdb>=0.7.1",
        "trycast>=1.0.0",
        "pydash>=7.0.3",
        "cachetools>=5.3.0",
        "pydantic==1.10.17",
        "aiocache>=0.12.1",
        "backoff>=2.2.1",
        "asyncstdlib>=3.10.7"
    ]
)



from typing import Dict, Any, List, Optional

from spotlight.api.data_rule.__util import (
    _get_data_rule_request_info,
    _get_data_rules_request_info,
    _create_data_rule_request_info,
    _update_data_rule_request_info,
    _delete_data_rule_request_info,
)
from spotlight.api.data_rule.model import DataRuleRequest
from spotlight.core.common.decorators import data_request
from spotlight.core.common.requests import (
    __get_request,
    __post_request,
    __put_request,
    __delete_request,
)


@data_request()
def get_data_rule(id: str) -> Dict[str, Any]:
    """
    Get data rule by ID.

    Args:
        id (str): Data rule ID

    Returns:
        Dict[str, Any]: Data rule response
    """
    request_info = _get_data_rule_request_info(id)
    return __get_request(**request_info)


@data_request()
def get_data_rules(override_find_all: Optional[bool] = False) -> List[Dict[str, Any]]:
    """
    Get all data rules.

    Args:
        override_find_all (Optional[bool]): If True, overrides the default limitation to fetch all results. This parameter is restricted to admin use only. Non-admin users will receive a 'Forbidden' exception if they attempt to use it. Defaults to False.

    Returns:
        List[Dict[str, Any]]: List of data rule responses
    """
    request_info = _get_data_rules_request_info(override_find_all)
    return __get_request(**request_info)


@data_request()
def create_data_rule(request: DataRuleRequest) -> Dict[str, Any]:
    """
    Create data rule.

    Args:
        request (DataRuleRequest): Data rule request

    Returns:
        Dict[str, Any]: Data rule response
    """
    request_info = _create_data_rule_request_info(request)
    return __put_request(**request_info)


@data_request()
def update_data_rule(id: str, request: DataRuleRequest) -> Dict[str, Any]:
    """
    Update data rule.

    Args:
        id (str): Data rule ID
        request (DataRuleRequest): Data rule request

    Returns:
        Dict[str, Any]: Data rule response
    """
    request_info = _update_data_rule_request_info(id, request)
    return __post_request(**request_info)


@data_request(processor=lambda response: None)
def delete_data_rule(id: str) -> None:
    """
    Delete data rule by ID.

    Args:
        id (str): Data rule ID

    Returns:
        None
    """
    request_info = _delete_data_rule_request_info(id)
    return __delete_request(**request_info)



from spotlight.core.common.date.function import (
    date_to_timestamp,
    datetime_to_timestamp,
    datetime_to_utc,
    date_to_datetime,
    standardize_date_or_datetime,
    timestamp_to_datetime,
)



"""
Common decorators.
"""

from spotlight.core.common.decorators import timeit
from spotlight.core.common.decorators.authorization import (
    authenticated_request,
    async_authenticated_request,
)
from spotlight.core.common.decorators.data import data_request, async_data_request
from spotlight.core.common.decorators.timeit import timeit



import logging
from functools import reduce
from typing import Any, Callable, List

from requests import Response

from spotlight.core.common.errors import ApiError

logger = logging.getLogger(__name__)


def __process_response(
    response: Response,
    processor: Callable[[Response], Any],
    transformers: List[Callable[[Any], Any]],
):
    if response.status_code != 200 | response.status_code != 201:
        error = ApiError(
            f"Request to {response.url} failed with status code {response.status_code}: {response.text}"
        )
        logger.error(error)
        raise error

    data = processor(response)
    return reduce(lambda res, fxn: fxn(res), transformers, data)



from typing import Dict, Any, List

from spotlight.api.dataset.model import SearchRequest
from spotlight.api.dataset.view.__util import (
    _get_dataset_view_request_info,
    _get_dataset_views_request_info,
    _search_dataset_views_request_info,
)
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import (
    __async_get_request,
    __async_post_request,
)


@async_data_request()
async def async_get_dataset_view(id: str) -> Dict[str, Any]:
    """
    Asynchronously get dataset view by ID.

    Args:
        id (str): Dataset ID

    Returns:
        Dict[str, Any]: Dataset view response
    """
    request_info = _get_dataset_view_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request()
async def async_get_dataset_views() -> List[Dict[str, Any]]:
    """
    Asynchronously get all dataset views.

    Returns:
        List[Dict[str, Any]]: List of dataset view response
    """
    request_info = _get_dataset_views_request_info()
    return await __async_get_request(**request_info)


@async_data_request()
async def async_search_dataset_views(
    request: SearchRequest,
) -> List[Dict[str, Any]]:
    """
    Asynchronously search all dataset views.

    Args:
        request (SearchRequest): Dataset view search request

    Returns:
        List[Dict[str, Any]]: List of dataset view response
    """
    request_info = _search_dataset_views_request_info(request)
    return await __async_post_request(**request_info)



from datetime import date
from typing import Dict, Any, List

from spotlight.core.common.enum import AssetClass


def _sdr_metadata_request_info(
    asset_class: AssetClass, from_date: date, to_date: date, reprocess: bool = True
) -> Dict[str, Any]:
    endpoint = f"sdr-source/sdr/{asset_class.lower()}/metadata"
    params = {
        "from": from_date.isoformat(),
        "to": to_date.isoformat(),
        "reprocess": str(reprocess).lower(),
    }
    return {"endpoint": endpoint, "params": params}


def _sdr_upload_request_info(metadata: List[Dict[str, Any]]) -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/upload"
    return {"endpoint": endpoint, "data": metadata}


def _sdr_download_request_info(metadata: Dict[str, Any]) -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/download"
    return {"endpoint": endpoint, "json": metadata}


def _sdr_download_json_request_info(metadata: Dict[str, Any]) -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/download?json"
    return {"endpoint": endpoint, "json": metadata}


def _sdr_cache_clear_file_request_info() -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/cache/clear?file"
    return {"endpoint": endpoint}


def _sdr_cache_clear_record_request_info() -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/cache/clear?record"
    return {"endpoint": endpoint}


def _sdr_cache_lookup_request_info(metadata: Dict[str, Any]) -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/cache/lookup"
    return {"endpoint": endpoint, "json": metadata}


def _sdr_status_request_info() -> Dict[str, Any]:
    endpoint = "sdr-source/sdr/status"
    return {"endpoint": endpoint}



"""
Authorization decorators.
"""

import functools
from typing import Any, Callable, Optional

from aiocache.decorators import cached
from requests import Response

from spotlight.api.auth import async_login, async_refresh_token
from spotlight.core.common.config import EnvironmentConfig
from spotlight.core.common.decorators.authorization.__util import (
    _access_token,
    _handle_auth_response,
    _validate_headers,
)


def async_authenticated_request(
    request: Callable = None, *, env_config: Optional[EnvironmentConfig] = None
) -> Callable:
    """
    Decorator for requests that need to be authenticated

    Args:
          request (Callable): Method that needs auth headers injected
          env_config (Optional[EnvironmentConfig]): Optional environment config
    Returns:
        Callable: Wrapped function
    """

    @functools.wraps(request)
    def decorator(f):
        @functools.wraps(f)
        async def wrap(*args, **kwargs) -> Any:
            """
            Wrapper for the decorated function

            Args:
                *args: args for the function
                **kwargs: keyword args for the function

            Returns:
                Any: The output of the wrapped function
            """
            config = env_config if env_config is not None else EnvironmentConfig()
            headers = kwargs.get("headers", {})

            if config.auth_config.can_authenticate():
                auth_headers = await _authenticate(
                    username=config.auth_config.username,
                    password=config.auth_config.password,
                    access_token=config.auth_config.access_token,
                    refresh_token=config.auth_config.refresh_token,
                )
                auth_headers.update(
                    headers
                )  # override authentication header if already provided
                kwargs["headers"] = auth_headers

            return await _make_request(f, config, *args, **kwargs)

        return wrap

    if request is None:
        return decorator
    return decorator(request)


@cached(ttl=1740)
async def _authenticate(
    username: Optional[str],
    password: Optional[str],
    access_token: Optional[str],
    refresh_token: Optional[str],
) -> dict:
    """
    Asynchronously authenticate user and create user authentication headers.

    Args:
        username (Optional[str]): Spotlight username
        password (Optional[str]): Spotlight password
        access_token (Optional[str]): Spotlight access token
        refresh_token (Optional[str]): Spotlight refresh token

    Returns:
        dict: Auth headers for a request
    """
    auth_headers = {}
    if username and password:
        auth_headers = await _sign_in(username=username, password=password)
    elif access_token:
        auth_headers = _access_token(access_token)
    elif refresh_token:
        auth_headers = await _refresh_token(refresh_token)
    return auth_headers


async def _sign_in(username: str, password: str) -> dict:
    auth_response = await async_login(username, password)
    return _handle_auth_response(auth_response, error_msg="Failed to sign-in the user")


async def _refresh_token(refresh_token: Optional[str] = None) -> dict:
    auth_response = await async_refresh_token(refresh_token)
    return _handle_auth_response(
        auth_response, error_msg="Failed to refresh the user's token"
    )


async def _make_request(func, config: EnvironmentConfig, *args, **kwargs) -> Response:
    _validate_headers(**kwargs)

    response: Response = await func(*args, **kwargs)
    if response.status_code == 401 and config.auth_config.refresh_token:
        auth_headers = await _refresh_token(config.auth_config.refresh_token)
        kwargs["headers"].update(auth_headers)  # override authentication header
        response: Response = await func(*args, **kwargs)

    return response



from typing import Dict, Any, List

from spotlight.api.dataset.abstract.__util import (
    _get_abstract_dataset_request_info,
    _get_abstract_datasets_request_info,
    _search_abstract_datasets_request_info,
    _create_abstract_dataset_request_info,
    _update_abstract_dataset_request_info,
    _delete_abstract_dataset_request_info,
)
from spotlight.api.dataset.abstract.model import AbstractDatasetRequest
from spotlight.api.dataset.model import SearchRequest
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import (
    __async_get_request,
    __async_post_request,
    __async_put_request,
    __async_delete_request,
)


@async_data_request()
async def async_get_abstract_dataset(id: str) -> Dict[str, Any]:
    """
    Asynchronously get abstract dataset by ID.

    Args:
        id (str): Abstract dataset ID

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _get_abstract_dataset_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request()
async def async_get_abstract_datasets() -> List[Dict[str, Any]]:
    """
    Asynchronously get all abstract datasets.

    Returns:
        List[Dict[str, Any]]: List of abstract dataset response
    """
    request_info = _get_abstract_datasets_request_info()
    return await __async_get_request(**request_info)


@async_data_request()
async def async_search_abstract_datasets(
    request: SearchRequest,
) -> List[Dict[str, Any]]:
    """
    Asynchronously search all abstract datasets.

    Args:
        request (SearchRequest): Abstract dataset search request

    Returns:
        List[Dict[str, Any]]: List of abstract dataset response
    """
    request_info = _search_abstract_datasets_request_info(request)
    return await __async_post_request(**request_info)


@async_data_request()
async def async_create_abstract_dataset(
    request: AbstractDatasetRequest,
) -> Dict[str, Any]:
    """
    Asynchronously create abstract dataset.

    Args:
        request (AbstractDatasetRequest): Abstract dataset request

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _create_abstract_dataset_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request()
async def async_update_abstract_dataset(
    id: str, request: AbstractDatasetRequest
) -> Dict[str, Any]:
    """
    Asynchronously update abstract dataset.

    Args:
        id (str): Abstract dataset ID
        request (AbstractDatasetRequest): Abstract dataset request

    Returns:
        Dict[str, Any]: Abstract dataset response
    """
    request_info = _update_abstract_dataset_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(processor=lambda response: None)
async def async_delete_abstract_dataset(id: str) -> None:
    """
    Asynchronously delete abstract dataset by ID.

    Args:
        id (str): Abstract dataset ID

    Returns:
        None
    """
    request_info = _delete_abstract_dataset_request_info(id)
    return await __async_delete_request(**request_info)



from datetime import datetime
from typing import Optional, List

from pydantic import Field

from spotlight.api.data.model import WhereClause
from spotlight.core.common.base import Base
from spotlight.core.common.enum import IntervalType, NotificationType


class AlertRequest(Base):
    display_name: str
    description: Optional[str] = Field(default=None)
    data_rule_id: str
    interval_type: IntervalType
    interval: int
    interval_start: Optional[datetime] = Field(default=None)
    notification_type: NotificationType
    notification_source: List[str]


class AlertResponse(Base):
    id: str
    display_name: str
    description: Optional[str]
    data_rule_id: str
    interval_type: IntervalType
    interval: int
    interval_start: Optional[datetime]
    notification_type: NotificationType
    notification_source: List[str]
    created_by: str
    created_at: datetime
    updated_by: Optional[str]
    updated_at: Optional[int]


class AlertSignal(Base):
    alert_id: str
    alert_name: str
    alert_description: Optional[str]
    notification_type: NotificationType
    notification_source: List[str]
    interval_type: IntervalType
    interval: int
    interval_start_time: datetime
    window_start_time: datetime
    window_end_time: datetime
    where_clause: WhereClause
    dataset_id: str



from typing import Dict, Any, List

from spotlight.api.dataset.__util import (
    _get_dataset_request_info,
    _get_datasets_request_info,
    _search_datasets_request_info,
    _create_dataset_request_info,
    _update_dataset_request_info,
    _delete_dataset_request_info,
)
from spotlight.api.dataset.model import DatasetRequest
from spotlight.api.dataset.model import SearchRequest
from spotlight.core.common.decorators import async_data_request
from spotlight.core.common.requests import (
    __async_get_request,
    __async_post_request,
    __async_put_request,
    __async_delete_request,
)


@async_data_request()
async def async_get_dataset(id: str) -> Dict[str, Any]:
    """
    Asynchronously get dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _get_dataset_request_info(id)
    return await __async_get_request(**request_info)


@async_data_request()
async def async_get_datasets() -> List[Dict[str, Any]]:
    """
    Asynchronously get all datasets.

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _get_datasets_request_info()
    return await __async_get_request(**request_info)


@async_data_request()
async def async_search_datasets(request: SearchRequest) -> List[Dict[str, Any]]:
    """
    Asynchronously search all datasets.

    Args:
        request (SearchRequest): Dataset search request

    Returns:
        List[Dict[str, Any]]: List of dataset response
    """
    request_info = _search_datasets_request_info(request)
    return await __async_post_request(**request_info)


@async_data_request()
async def async_create_dataset(request: DatasetRequest) -> Dict[str, Any]:
    """
    Asynchronously create dataset.

    Args:
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _create_dataset_request_info(request)
    return await __async_put_request(**request_info)


@async_data_request()
async def async_update_dataset(id: str, request: DatasetRequest) -> Dict[str, Any]:
    """
    Asynchronously update dataset.

    Args:
        id (str): Dataset ID
        request (DatasetRequest): Dataset request

    Returns:
        Dict[str, Any]: Dataset response
    """
    request_info = _update_dataset_request_info(id, request)
    return await __async_post_request(**request_info)


@async_data_request(processor=lambda response: None)
async def async_delete_dataset(id: str) -> None:
    """
    Asynchronously delete dataset by ID.

    Args:
        id (str): Dataset ID

    Returns:
        None
    """
    request_info = _delete_dataset_request_info(id)
    return await __async_delete_request(**request_info)
